/*
 * Copyright (C) 2017 Open Source Robotics Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

#include <gtest/gtest.h>
#include <string>

#include <ignition/common/Console.hh>

#include "ignition/gui/Iface.hh"
#include "ignition/gui/VariablePill.hh"
#include "ignition/gui/VariablePillContainer.hh"

using namespace ignition;
using namespace gui;

/////////////////////////////////////////////////
TEST(VariablePillContainerTest, VariablePillBasicOperations)
{
  setVerbosity(4);
  EXPECT_TRUE(initApp());

  // create container
  VariablePillContainer *container01 = new VariablePillContainer(nullptr);
  ASSERT_NE(nullptr, container01);
  EXPECT_EQ(0u, container01->VariablePillCount());

  // create variable pills
  VariablePill *var01 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var01);
  var01->SetText("pill0");
  EXPECT_EQ(0u, var01->VariablePillCount());
  EXPECT_EQ(0u, var01->Id());
  EXPECT_EQ("variable0", var01->Name());
  EXPECT_EQ("pill0", var01->Text());
  EXPECT_FALSE(var01->IsSelected());
  var01->SetSelected(true);
  EXPECT_TRUE(var01->IsSelected());
  var01->SetSelected(false);
  EXPECT_FALSE(var01->IsSelected());

  VariablePill *var02 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var02);
  var02->SetText("pill1");
  EXPECT_EQ(0u, var02->VariablePillCount());
  EXPECT_EQ(1u, var02->Id());
  EXPECT_EQ("variable1", var02->Name());
  EXPECT_EQ("pill1", var02->Text());
  var02->SetName("newVariable1");
  EXPECT_EQ("newVariable1", var02->Name());

  var02->SetParent(var01);
  EXPECT_EQ(nullptr, var01->Parent());
  EXPECT_EQ(var01, var02->Parent());

  var01->SetContainer(container01);
  EXPECT_EQ(container01, var01->Container());

  EXPECT_TRUE(var01->ContainsPoint(ignition::math::Vector2i()));
  EXPECT_FALSE(var01->ContainsPoint(ignition::math::Vector2i(1000, 1000)));

  EXPECT_TRUE(stop());
}

/////////////////////////////////////////////////
TEST(VariablePillContainerTest, VariablePillEvents)
{
  setVerbosity(4);
  EXPECT_TRUE(initApp());

  // create container
  auto container01 = new VariablePillContainer(nullptr);
  ASSERT_NE(nullptr, container01);
  EXPECT_EQ(0u, container01->VariablePillCount());

  // create variable pills
  VariablePill *var01 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var01);

  VariablePill *var02 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var02);

  // First, we have a container with two variables. We're going to simulate
  // dragging one of the variables (var01) into the other (var02). At that
  // point we'll have a single multi-variable pill.
  container01->AddVariablePill(var01);
  container01->AddVariablePill(var02);
  QCoreApplication::processEvents();
  container01->show();
  QCoreApplication::processEvents();
  EXPECT_EQ(2u, container01->VariablePillCount());
  EXPECT_EQ(0u, var01->VariablePillCount());
  EXPECT_EQ(0u, var02->VariablePillCount());

  // Check the container begins at 0, 0 on Linux and higher on OSX
  auto initialContainerX = container01->pos().x();
  auto initialContainerY = container01->pos().y();
#if !defined(__APPLE__)
  EXPECT_EQ(0, initialContainerX);
  EXPECT_EQ(0, initialContainerY);
#else
  EXPECT_GT(initialContainerX, 0);
  EXPECT_GT(initialContainerY, 0);
#endif

  // Check both pills have the same size
  EXPECT_EQ(var01->width(), var02->width());
  EXPECT_EQ(var01->height(), var02->height());

  // Get the pill's center in its local frame
  QPoint varCenter(var01->width() * 0.5, var01->height() * 0.5);

  // Get the position of the pills' centers in the global frame
  auto var01Global = var01->mapToGlobal(varCenter);
  auto var02Global = var02->mapToGlobal(varCenter);

  // Check both pills are on the same vertical position
  EXPECT_EQ(var01Global.y(), var02Global.y());

  // And var02 is on the right of var01
  EXPECT_LT(var01Global.x(), var02Global.x());

  // Mouse-press the center of var01
  auto mousePressEvent = new QMouseEvent(QEvent::MouseButtonPress,
    varCenter, Qt::LeftButton, Qt::NoButton, Qt::NoModifier);
  QCoreApplication::postEvent(var01, mousePressEvent);
  QCoreApplication::processEvents();

  // Check the variables are still in the same place
  EXPECT_EQ(var01Global, var01->mapToGlobal(varCenter));
  EXPECT_EQ(var02Global, var02->mapToGlobal(varCenter));

  // Drag the mouse 1px, this moves the container locally, but not on Jenkins!
  auto mouseLocalPos = varCenter;
  auto mouseGlobalPos = var01Global;
  mouseGlobalPos.setX(mouseGlobalPos.x() + 1);
  mouseLocalPos.setX(mouseLocalPos.x() + 1);

  bool moved = false;
  QTimer::singleShot(50, [mouseLocalPos, mouseGlobalPos, &var01, &moved]
  {
    auto mouseMoveEvent = new QMouseEvent(QEvent::MouseMove,
      mouseLocalPos, mouseGlobalPos, Qt::LeftButton, Qt::LeftButton,
      Qt::NoModifier);
    QCoreApplication::postEvent(var01, mouseMoveEvent);
    QCoreApplication::processEvents();
    moved = true;
  });

  while (!moved)
  {
    QCoreApplication::processEvents();
  }
  EXPECT_TRUE(moved);

  // Check the variables are still the same size
  EXPECT_EQ(varCenter.x(), static_cast<int>(var02->width() * 0.5));
  EXPECT_EQ(varCenter.y(), static_cast<int>(var02->height() * 0.5));

  // \fixme Locally, the container moves to another place on the screen
  // On Jenkins it doesn't and the mouse never enters var02
  bool containerMoved = false;
  if (container01->pos().x() > initialContainerX ||
      container01->pos().y() > initialContainerY)
  {
    containerMoved = true;

    // If the container moves, the variables move too
    if (container01->pos().x() > initialContainerX)
    {
      EXPECT_LT(var01Global.x(), var01->mapToGlobal(varCenter).x());
    }
    if (container01->pos().y() > initialContainerY)
    {
      EXPECT_LT(var02Global.y(), var02->mapToGlobal(varCenter).y());
    }
  }
  else
  {
    igndbg << "Container didn't move" << std::endl;

    // If the container didn't move, the variables are the same
    EXPECT_EQ(var01Global.x(), var01->mapToGlobal(varCenter).x());
    EXPECT_EQ(var02Global.y(), var02->mapToGlobal(varCenter).y());
  }

  // Store their new global poses
  var01Global = var01->mapToGlobal(varCenter);
  var02Global = var02->mapToGlobal(varCenter);

  // Adjust the mouse position
  mouseGlobalPos = var01Global;
  mouseGlobalPos.setX(mouseGlobalPos.x() + 1);

  // Now keep dragging until the center of var02
  int created = 0;
  int triggered = 0;
  unsigned int diff = var02Global.x() - mouseGlobalPos.x();
  for (unsigned int i = 0; i < diff; ++i)
  {
    created++;
    QTimer::singleShot(50, [&mouseLocalPos, &mouseGlobalPos, &triggered,
                            &created, &var01, &var02, i, diff, varCenter]
    {
      // On the last move, also trigger release
      if (i == diff - 1)
      {
        created++;
        // Release.
        QTimer::singleShot(300, [varCenter, mouseGlobalPos, &var02, &triggered]
        {
          triggered++;

          auto mouseReleaseEvent = new QMouseEvent(QEvent::MouseButtonRelease,
              varCenter, mouseGlobalPos, Qt::LeftButton,
              Qt::LeftButton, Qt::NoModifier);
          QCoreApplication::postEvent(var02, mouseReleaseEvent);
          QCoreApplication::processEvents();
        });
      }
      triggered++;

      // Compute the next x pos to move the mouse cursor to.
      mouseLocalPos.setX(mouseLocalPos.x() + 1);
      mouseGlobalPos.setX(mouseGlobalPos.x() + 1);

      auto mouseMoveEvent = new QMouseEvent(QEvent::MouseMove,
          mouseLocalPos, mouseGlobalPos, Qt::LeftButton, Qt::LeftButton,
          Qt::NoModifier);
      QCoreApplication::postEvent(var01, mouseMoveEvent);
      QCoreApplication::processEvents();
    });
  }

  while (triggered < created)
  {
    QCoreApplication::processEvents();
  }
  EXPECT_EQ(triggered, created);

  // Then, a container with one multi-variable pill.
  EXPECT_EQ(2u, container01->VariablePillCount());
  EXPECT_EQ(0u, var01->VariablePillCount());
  // \fixme When the container doesn't move, the test fails, this happens on
  // Jenkins
  if (containerMoved)
  {
    EXPECT_EQ(1u, var02->VariablePillCount());
  }

  EXPECT_TRUE(stop());
}

/////////////////////////////////////////////////
TEST(VariablePillContainerTest, VariablePillContainerBasicOperations)
{
  setVerbosity(4);
  EXPECT_TRUE(initApp());

  // create container
  VariablePillContainer *container01 = new VariablePillContainer(nullptr);
  ASSERT_NE(nullptr, container01);
  EXPECT_EQ(0u, container01->VariablePillCount());
  container01->SetText("aContainer");
  EXPECT_EQ("aContainer", container01->Text());

  auto id0 = container01->AddVariablePill("var0");
  container01->SetVariablePillLabel(id0, "label0");
  auto var0 = container01->VariablePill(id0);
  ASSERT_NE(nullptr, var0);
  EXPECT_EQ("label0", var0->Text());

  auto id1 = container01->AddVariablePill("var1");
  auto var1 = container01->VariablePill(id1);
  ASSERT_NE(nullptr, var1);

  EXPECT_FALSE(var0->IsSelected());
  EXPECT_FALSE(var1->IsSelected());
  container01->SetSelected(nullptr);
  EXPECT_FALSE(var0->IsSelected());
  EXPECT_FALSE(var1->IsSelected());

  container01->SetSelected(var0);
  EXPECT_TRUE(var0->IsSelected());
  EXPECT_FALSE(var1->IsSelected());
  container01->SetSelected(var1);
  EXPECT_FALSE(var0->IsSelected());
  EXPECT_TRUE(var1->IsSelected());

  EXPECT_EQ(2u, container01->VariablePillCount());
  container01->RemoveVariablePill(id0);
  container01->RemoveVariablePill(id1);
  EXPECT_EQ(0u, container01->VariablePillCount());

  EXPECT_TRUE(stop());
}

/////////////////////////////////////////////////
TEST(VariablePillContainerTest, AddRemoveVariable)
{
  setVerbosity(4);
  EXPECT_TRUE(initApp());

  // create container
  VariablePillContainer *container01 = new VariablePillContainer(nullptr);
  ASSERT_NE(nullptr, container01);
  EXPECT_EQ(0u, container01->VariablePillCount());

  // create variable pills
  VariablePill *var01 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var01);
  EXPECT_EQ(0u, var01->VariablePillCount());

  VariablePill *var02 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var02);
  EXPECT_EQ(0u, var02->VariablePillCount());

  VariablePill *var03 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var03);
  EXPECT_EQ(0u, var03->VariablePillCount());

  VariablePill *var04 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var04);
  EXPECT_EQ(0u, var04->VariablePillCount());

  VariablePill *var05 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var05);
  EXPECT_EQ(0u, var05->VariablePillCount());
  var05->SetName("var05");

  // add variable to container
  container01->AddVariablePill(var01);
  EXPECT_EQ(1u, container01->VariablePillCount());
  EXPECT_EQ(container01, var01->Container());
  EXPECT_EQ(nullptr, var01->Parent());

  // add another variable to container
  container01->AddVariablePill(var02);
  EXPECT_EQ(2u, container01->VariablePillCount());
  EXPECT_EQ(container01, var02->Container());
  EXPECT_EQ(nullptr, var02->Parent());

  // add an invalid variable pill
  var02->AddVariablePill(nullptr);
  EXPECT_EQ(2u, container01->VariablePillCount());

  // add variable to another variable - verify that containers can hold
  // multi-variables and report correct variable count
  var02->AddVariablePill(var03);
  EXPECT_EQ(3u, container01->VariablePillCount());
  EXPECT_EQ(container01, var03->Container());
  EXPECT_EQ(var02, var03->Parent());

  EXPECT_EQ(1u, var02->VariablePills().size());

  // make another multi-variable
  var01->AddVariablePill(var04);
  EXPECT_EQ(4u, container01->VariablePillCount());
  EXPECT_EQ(container01, var04->Container());
  EXPECT_EQ(var01, var04->Parent());

  // add variable to a variable which is within a multi-variable
  var04->AddVariablePill(var05);
  EXPECT_EQ(5u, container01->VariablePillCount());
  EXPECT_EQ(container01, var05->Container());
  EXPECT_EQ(var01, var05->Parent());

  EXPECT_EQ(2u, var01->VariablePills().size());

  EXPECT_EQ(nullptr, var04->VariablePillByName("___wrong_name___"));
  EXPECT_EQ(nullptr, var04->VariablePillByName("var05"));
  EXPECT_EQ(var05, var01->VariablePillByName("var05"));
  EXPECT_EQ(nullptr, var02->VariablePillByName("var05"));
  EXPECT_EQ(var05, var05->VariablePillByName("var05"));

  // remove variable
  container01->RemoveVariablePill(var01);
  EXPECT_EQ(4u, container01->VariablePillCount());
  EXPECT_EQ(nullptr, var01->Parent());

  // remove already removed variable
  container01->RemoveVariablePill(var01);
  EXPECT_EQ(4u, container01->VariablePillCount());

  // remove child of a multi-variable - check that the parent is not affected
  container01->RemoveVariablePill(var03);
  EXPECT_EQ(3u, container01->VariablePillCount());
  EXPECT_EQ(container01, var02->Container());
  EXPECT_EQ(nullptr, var02->Parent());
  EXPECT_EQ(nullptr, var03->Parent());

  // remove multi-variable - check that it doesn't remove the child variable
  container01->RemoveVariablePill(var04);
  EXPECT_EQ(2u, container01->VariablePillCount());
  ASSERT_EQ(nullptr, var05->Parent());
  EXPECT_EQ(container01, var05->Container());
  EXPECT_EQ(nullptr, var04->Parent());
  EXPECT_EQ(nullptr, var05->Parent());

  // remove remaining variables
  container01->RemoveVariablePill(var02);
  EXPECT_EQ(1u, container01->VariablePillCount());
  EXPECT_EQ(nullptr, var02->Parent());

  container01->RemoveVariablePill(var05);
  EXPECT_EQ(0u, container01->VariablePillCount());
  EXPECT_EQ(nullptr, var05->Parent());

  delete var05;
  delete var04;
  delete var03;
  delete var02;
  delete var01;
  delete container01;

  EXPECT_TRUE(stop());
}

/////////////////////////////////////////////////
TEST(VariablePillContainerTest, MaxSize)
{
  setVerbosity(4);
  EXPECT_TRUE(initApp());

  // create new container
  VariablePillContainer *container =  new VariablePillContainer(nullptr);
  ASSERT_NE(nullptr, container);

  // set text
  std::string containerName = "container_test";
  container->SetText(containerName);
  EXPECT_EQ(containerName, container->Text());

  // set max size
  EXPECT_EQ(container->MaxSize(), -1);
  int maxSize = 1;
  container->SetMaxSize(maxSize);
  EXPECT_EQ(maxSize, container->MaxSize());

  // create variable pills
  VariablePill *var01 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var01);
  EXPECT_EQ(0u, var01->VariablePillCount());

  VariablePill *var02 = new VariablePill(nullptr);
  ASSERT_NE(nullptr, var02);
  EXPECT_EQ(0u, var02->VariablePillCount());

  // add variable to container - max size reached
  container->AddVariablePill(var01);
  EXPECT_EQ(1u, container->VariablePillCount());
  EXPECT_EQ(container, var01->Container());
  EXPECT_EQ(nullptr, var01->Parent());

  // verify no more variables can be added
  container->AddVariablePill(var02);
  EXPECT_EQ(1u, container->VariablePillCount());
  ASSERT_EQ(nullptr, var02->Container());
  EXPECT_EQ(nullptr, var02->Parent());

  var01->AddVariablePill(var02);
  EXPECT_EQ(1u, container->VariablePillCount());
  EXPECT_EQ(nullptr, var02->Container());
  EXPECT_EQ(nullptr, var02->Parent());

  // remove variable and verify we can add a different variable now
  container->RemoveVariablePill(var01);
  EXPECT_EQ(0u, container->VariablePillCount());
  EXPECT_EQ(nullptr, var01->Container());
  EXPECT_EQ(nullptr, var01->Parent());

  container->AddVariablePill(var02);
  EXPECT_EQ(1u, container->VariablePillCount());
  EXPECT_EQ(container, var02->Container());
  EXPECT_EQ(nullptr, var02->Parent());

  delete var02;
  delete var01;
  delete container;

  EXPECT_TRUE(stop());
}
